from __future__ import annotations
from functools import cached_property
import json
import os
import time
from typing import Any, Dict, List
from datetime import datetime
import polars as pl
from web3.constants import ADDRESS_ZERO
from web3.contract import Contract
import numpy as np

from web3cat.fetcher.erc20_metas import ERC20Meta
from web3cat.fetcher.events import Event
from web3cat.data.core import DataCore

RESOLVER_MAPPING = {"weth": "eth", "wbtc": "btc"}


class ChainlinkUSDData(DataCore):
    """
    Chainlink data for a specific token.
    """

    _token: str

    def __init__(
        self, token: str, start: int | datetime, end: int | datetime, **kwargs
    ):
        super().__init__(start, end, **kwargs)
        self._token = token

    @cached_property
    def meta(self) -> ERC20Meta:
        """
        Metadata for tokens (like name, symbol and decimals)
        """
        return self._erc20_metas_service.get(self._token)

    @cached_property
    def oracle_proxy_contract(self) -> Contract:
        """
        A reference to Chainlink oracle proxy contract
        """
        current_folder = os.path.realpath(os.path.dirname(__file__))
        with open(
            f"{current_folder}/oracle_proxy.abi.json", "r", encoding="utf-8"
        ) as f:
            abi = json.load(f)
        oracle_address = self._resolve_chainlink_address(self.meta.symbol.lower())
        if oracle_address is None:
            raise LookupError(
                f"Chainlink oracle for token `{self.meta.symbol.lower()}` "
                f"on chain with id `{self._blocks_service.chain_id}` is not found"
            )

        return self.w3.eth.contract(
            address=self.w3.to_checksum_address(oracle_address), abi=abi
        )

    @cached_property
    def oracle_aggregator_contract(self) -> Contract:
        """
        A reference to Chainlink oracle aggregator contract
        """

        current_folder = os.path.realpath(os.path.dirname(__file__))
        with open(
            f"{current_folder}/oracle_aggregator.abi.json", "r", encoding="utf-8"
        ) as f:
            abi = json.load(f)
        oracle_address = self._calls_service.get_call(
            self.oracle_proxy_contract.functions.aggregator(),
            self.to_block_number,
        ).response
        return self.w3.eth.contract(
            address=self.w3.to_checksum_address(oracle_address),
            abi=abi,
        )

    @cached_property
    def oracle_decimals(self) -> int:
        """
        Chainlink oracle decimals
        """
        return int(
            self._calls_service.get_call(
                self.oracle_proxy_contract.functions.decimals(),
                self.to_block_number,
            ).response
        )

    @cached_property
    def initial_price(self) -> np.float64:
        """
        Oracle price at from_block
        """
        price = self._calls_service.get_call(
            self.oracle_aggregator_contract.functions.latestRoundData(),
            self.from_block_number,
        ).response[1]
        return price / 10**self.oracle_decimals

    @cached_property
    def index(self) -> Dict[str, Any]:
        """
        Oracles index
        """
        current_folder = os.path.realpath(os.path.dirname(__file__))
        with open(f"{current_folder}/oracles.json", "r", encoding="utf-8") as f:
            return json.load(f)

    def prices(self, timepoints: List[int | datetime]) -> pl.DataFrame:
        """
        Chainlink price feed.

        Arguments:
            timepoints: A list of timepoints for prices
        """

        blocks = self._resolve_timepoints(sorted(timepoints))
        updates = self.updates[["block_number", "price"]].to_dicts()
        i = 0
        price_list = []
        while len(updates) > 0 and blocks[i].number < updates[0]["block_number"]:
            price_list.append(self.initial_price)
            i += 1
            if i == len(blocks):
                break
        j = 0
        for b in blocks[i:]:
            while j < len(updates) and updates[j]["block_number"] <= b.number:
                j += 1
            # now ts < updates[j]
            idx = max([j - 1, 0])
            price_list.append(updates[idx]["price"])
        out = [
            {
                "timestamp": b.timestamp,
                "date": datetime.fromtimestamp(b.timestamp),
                "block_number": b.number,
                "price": price,
            }
            for b, price in zip(blocks, price_list)
        ]
        return pl.DataFrame(
            out,
            {
                "timestamp": pl.UInt64,
                "date": pl.Datetime,
                "block_number": pl.UInt64,
                "price": pl.Float64,
            },
        )

    @cached_property
    def updates(self) -> pl.DataFrame:
        """
        A list of oracle updates
        """
        events: List[Event] = self._events_service.get_events(
            self.oracle_aggregator_contract.events.AnswerUpdated,
            self.from_block_number,
            self.to_block_number,
        )

        block_numbers = [e.block_number for e in events]
        blocks = self._blocks_service.get_blocks(block_numbers)
        timestamps = [b.timestamp for b in blocks]
        ts_index = dict(zip(block_numbers, timestamps))
        factor = 10**self.oracle_decimals
        return pl.DataFrame(
            [self._event_to_row(e, ts_index[e.block_number], factor) for e in events],
            {
                "timestamp": pl.UInt64,
                "date": pl.Datetime,
                "block_number": pl.UInt64,
                "transaction_hash": pl.Utf8,
                "log_index": pl.UInt64,
                "round": pl.UInt64,
                "updated_at": pl.UInt64,
                "price": pl.Float64,
            },
        ).sort(pl.col("timestamp"))

    def _event_to_row(self, ev: Event, ts: int, val_factor: int) -> Dict[str, Any]:
        return {
            "timestamp": ts,
            "date": datetime.fromtimestamp(ts),
            "block_number": ev.block_number,
            "transaction_hash": ev.transaction_hash,
            "log_index": ev.log_index,
            "round": ev.args["roundId"],
            "updated_at": ev.args["updatedAt"],
            "price": ev.args["current"] / val_factor,
        }

    def _resolve_chainlink_address(self, token: str) -> str | None:
        cid = str(self._blocks_service.chain_id)
        if not cid in self.index:
            return None
        oracles = self.index[cid]
        if token in oracles:
            return oracles[token]["address"]
        token = RESOLVER_MAPPING.get(token, token)
        if not token in oracles:
            return None
        return oracles[token]["address"]

    def _resolve_timestamps(self, timestamps: List[int | datetime]) -> List[int]:
        resolved = []
        for ts in timestamps:
            # resolve datetimes to timestamps
            if isinstance(ts, datetime):
                resolved.append(int(time.mktime(ts.timetuple())))
            else:
                resolved.append(ts)
        return resolved


class ChainlinkData(DataCore):
    """
    Chainlink price feed for tokens.

    Arguments:
        tokens: a list of ERC20 tokens
        start: start of a dataset
        end: end of a dataset
        kwargs: arguments for fetchers, see :class:`fetcher.core.Core`
    """

    _datas: Dict[str, ChainlinkUSDData]
    _kwargs: Dict[str, Any]

    def __init__(
        self,
        tokens: List[str],
        start: int | datetime,
        end: int | datetime,
        **kwargs,
    ):
        super().__init__(start, end, **kwargs)
        self._datas = {}
        self._kwargs = kwargs
        for token in tokens:
            self.add_token(token)

    def add_token(
        self,
        token: str,
        start: int | datetime | None = None,
        end: int | datetime | None = None,
        proxy: str | None = None,
    ):
        """
        Add token to prices dataset.

        Arguments:
            token: token to add
            start: start of the data
            end: end of the data
        """
        if token == ADDRESS_ZERO or token.upper() == "USD":
            return
        meta = self._erc20_metas_service.get(token)
        if proxy is None:
            self._datas[meta.address] = ChainlinkUSDData(
                token, start or self.start, end or self.end, **self._kwargs
            )
        else:
            proxy_meta = self._erc20_metas_service.get(token)
            self._datas[meta.address] = self._datas[proxy_meta.address]

    def get_data(self, token: str) -> ChainlinkUSDData:
        """
        Get chainlink USD data for token.

        Agruments:
            token: ERC20 token

        Returns:
            Chainlink USD data for token
        """

        meta = self._erc20_metas_service.get(token)
        if not meta.address in self._datas:
            self.add_token(token)
        return self._datas[meta.address]

    def prices(
        self, token0: str, token1: str, timepoints: List[int | datetime]
    ) -> pl.DataFrame:
        """
        Chainlink prices for pair of tokens.
        The token0 = "WETH", token1 = "USDC" would mean the regular WETH/USDC
        (that is how much dollars we give for 1 ETH)

        Dataframe with balances for addresses over time.

        Args:
            token0: The first token
            token1: The second token
            timepoints: A list of timepoints (see :class:`ERC20Data`)

        Returns:
            A Dataframe with fields

        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | Field                | Type                       | Description                                                                  |
        +======================+============================+==============================================================================+
        | ``timestamp``        | :attr:`numpy.int64`        | Timestamp for the snapshot of the balance                                    |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``date``             | :class:`datetime.datetime` | Date for the timestamp                                                       |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``block_number``     | :class:`int`               | Number of the block                                                          |
        +----------------------+----------------------------+------------------------------------------------------------------------------+
        | ``price``            | :attr:`numpy.float64`      | Relative price                                                               |
        +----------------------+----------------------------+------------------------------------------------------------------------------+

        """
        blocks = self._resolve_timepoints(timepoints)
        token0_is_usd = token0.upper() == "USD" or token0 == ADDRESS_ZERO
        token1_is_usd = token1.upper() == "USD" or token1 == ADDRESS_ZERO

        if token0_is_usd and token1_is_usd:
            prices = [1.0] * len(timepoints)
        elif token0_is_usd:
            data = self.get_data(token1)
            prices = [1.0 / p for p in data.prices(timepoints)["price"]]
        elif token1_is_usd:
            data = self.get_data(token0)
            prices = data.prices(timepoints)["price"]
        else:
            token0_data = self.get_data(token0)
            token1_data = self.get_data(token1)
            block_numbers = [b.number for b in blocks]
            prices0 = token0_data.prices(block_numbers)["price"].to_list()
            prices1 = token1_data.prices(block_numbers)["price"].to_list()
            prices = [p0 / p1 for p0, p1 in zip(prices0, prices1)]

        out = [
            {
                "timestamp": b.timestamp,
                "date": datetime.fromtimestamp(b.timestamp),
                "block_number": b.number,
                "price": p,
            }
            for b, p in zip(blocks, prices)
        ]
        df = pl.DataFrame(
            out,
            {
                "timestamp": pl.UInt64,
                "date": pl.Datetime,
                "block_number": pl.UInt64,
                "price": pl.Float64,
            },
        )
        return df



from __future__ import annotations
from itertools import product
import json
from typing import List
from web3.contract import ContractFunction
from web3cat.fetcher.calls.call import Call

from web3cat.fetcher.core import Core
from web3cat.fetcher.utils import calldata, json_response, print_progress
from web3cat.fetcher.calls.repo import CallsRepo


class CallsService(Core):
    """
    Service for making contract static calls.

    The sole purpose of this service is to cache web 3 calls and
    serve them on subsequent calls.

    **Request/Response flow**

    ::

                +---------------+                 +-------+ +-----------+
                | CallsService  |                 | Web3  | | CallsRepo |
                +---------------+                 +-------+ +-----------+
        ---------------  |                             |           |
        | Request call |-|                             |           |
        |--------------| |                             |           |
                         |                             |           |
                         | Find response               |           |
                         |---------------------------------------->|
                         |                             |           |
                         | If not found: call Web3     |           |
                         |---------------------------->|           |
                         |                             |           |
                         | Save response               |           |
                         |---------------------------------------->|
            -----------  |                             |           |
            | Response |-|                             |           |
            |----------| |                             |           |
                         |                             |           |


    Args:
        calls_repo: :class:`CallsRepo` instance
        kwargs: Args for the :class:`fetcher.core.Core`
    """

    _calls_repo: CallsRepo

    def __init__(self, calls_repo: CallsRepo, **kwargs):
        super().__init__(**kwargs)
        self._calls_repo = calls_repo

    @staticmethod
    def create(**kwargs) -> CallsService:
        """
        Create an instance of :class:`CallsService`

        Args:
            kwargs: Args for the :class:`fetcher.core.Core`

        Returns:
            An instance of :class:`CallsService`
        """
        calls_repo = CallsRepo(**kwargs)
        return CallsService(calls_repo, **kwargs)

    def get_call(
        self,
        call: ContractFunction,
        block_number: int,
    ) -> Call:
        """
        Make contract call specified by parameters.

        Args:
            call: :class:`web3.contract.ContractFunction` specifying contract, function
                  and call arguments.
            block_number: get call at this block

        Returns:
            A fetched :class:`fetcher.calls.Call`
        """

        data = calldata(call)
        calls = list(
            self._calls_repo.find(call.address, data, block_number, block_number + 1)
        )
        if len(calls) > 0:
            return calls[0]

        resp = json.loads(json_response(call.call(block_identifier=block_number)))
        call_item = Call(self.chain_id, call.address, data, block_number, resp)
        self._calls_repo.save([call_item])
        self._calls_repo.conn.commit()

        return call_item

    def get_calls(
        self,
        calls: List[ContractFunction],
        block_numbers: List[int],
    ) -> List[Call]:
        """
        Make a list of contract calls specified by parameters.

        Note:
            Each call is make for each block

        Args:
            calls: a list of :class:`web3.contract.ContractFunction` specifying contract, function
                  and call arguments.
            block_number: A list of blocks

        Returns:
            A fetched :class:`fetcher.calls.Call`
        """
        ids = set(
            f"{call.address.lower()}|{calldata(call)}|{block_number}"
            for call, block_number in product(calls, block_numbers)
        )
        calls_idx = {}
        for call in calls:
            calls_idx[f"{call.address.lower()}|{calldata(call)}"] = call
        idx = {}
        cached_calls = list(
            self._calls_repo.find_many(
                [(call.address, calldata(call)) for call in calls], block_numbers
            )
        )
        for call in cached_calls:
            call_id = f"{call.address.lower()}|{call.calldata}|{call.block_number}"
            idx[call_id] = call
            ids.remove(call_id)

        for i, call_id in enumerate(ids):
            print_progress(i, len(ids), f"Fetching {len(ids)} calls")
            address, cd, block_number = call_id.split("|")
            call = calls_idx[f"{address}|{cd}"]
            fetched_call = self.get_call(call, int(block_number))
            idx[call_id] = fetched_call
        if len(ids) > 0:
            print_progress(len(ids), len(ids), f"Fetching {len(ids)} calls")
        out = []

        for call in calls:
            for block_number in block_numbers:
                idx_id = f"{call.address.lower()}|{calldata(call)}|{block_number}"
                out.append(idx[idx_id])

        return out

    def clear_cache(self):
        """
        Delete all cached entries
        """
        self._calls_repo.purge()
        self._calls_repo.conn.commit()



from typing import Iterator, List, Tuple
from web3cat.fetcher.calls.call import Call
from web3cat.fetcher.core import Core


class CallsRepo(Core):
    """
    Reading and writing :class:`Call` to database.
    """

    def find(
        self,
        address: str,
        calldata: str,
        from_block: int = 0,
        to_block: int = 2**32 - 1,
    ) -> Iterator[Call]:
        """
        Find all calls in the database.

        Args:
            chain_id: Ethereum chain_id
            address: Contract address
            calldata: Ethereum calldata
            from_block: starting from this block (inclusive)
            to_block: ending with this block (non-inclusive)

        Returns:
            List of found calls
        """
        rows = self.conn.execute(
            "SELECT * FROM calls WHERE chain_id = ? AND address = ? "
            "AND calldata = ? AND block_number >= ? AND block_number < ?",
            [self.chain_id, address.lower(), calldata, from_block, to_block],
        )
        return (Call.from_row(r) for r in rows)

    def find_many(
        self,
        addresses_and_calldatas: List[Tuple[str, str]],
        blocks: List[int],
    ) -> Iterator[Call]:
        """
        Find all calls in the database.

        Args:
            addresses_and_calldatas: A list of addresses and calldatas
            blocks: A list of blocks

        Returns:
            A list of calls
        """
        if len(addresses_and_calldatas) == 0 or len(blocks) == 0:
            return []
        address_statement = " OR ".join(
            ["(address = ? AND calldata = ?)"] * len(addresses_and_calldatas)
        )
        address_args = [item.lower() for ac in addresses_and_calldatas for item in ac]
        blocks_statement = ",".join(["?"] * len(blocks))
        statement = (
            f"SELECT * FROM calls WHERE chain_id = ? AND ({address_statement}) "
            f"AND block_number IN ({blocks_statement})"
        )
        rows = self.conn.execute(
            statement,
            [self.chain_id, *address_args, *blocks],
        )
        return (Call.from_row(r) for r in rows)

    def save(self, calls: List[Call]):
        """
        Save a list of calls into the database.

        Args:
            calls: List of calls to save
        """
        rows = [e.to_row() for e in calls]
        self.conn.executemany(
            "INSERT INTO calls VALUES(?,?,?,?,?) ON CONFLICT DO NOTHING", rows
        )

    def purge(self):
        """
        Clear all database entries
        """
        cursor = self.conn.cursor()
        cursor.execute("DELETE FROM calls")



from __future__ import annotations
import json
from typing import Any, Dict, Tuple


class Block:
    """
    Ethereum block data
    """

    #: Ethereum chain_id
    chain_id: int
    #: Block number
    number: int
    #: Block timestamp
    timestamp: int

    def __init__(self, chain_id: int, number: int, timestamp: int):
        self.chain_id = chain_id
        self.number = number
        self.timestamp = timestamp

    @staticmethod
    def from_row(row: Tuple[int, int, int]) -> Block:
        """
        Deserialize from web3cat.database row

        Args:
            row: database row
        """
        return Block(*row)

    def to_row(self) -> Tuple[int, int, int]:
        """
        Serialize to database row

        Returns:
            database row
        """
        return (self.chain_id, self.number, self.timestamp)

    @staticmethod
    def from_dict(dct: Dict[str, Any]):
        """
        Create :class:`Block` from dict
        """
        return Block(
            chain_id=dct["chainId"],
            number=dct["number"],
            timestamp=dct["timestamp"],
        )

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert :class:`Block` to dict
        """
        return {
            "chainId": self.chain_id,
            "number": self.number,
            "timestamp": self.timestamp,
        }

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False

    def __repr__(self):
        return f"Block({json.dumps(self.to_dict())})"



from typing import List, Any
from datetime import datetime
import time
from dataclasses import dataclass
from bokeh.plotting import Figure
from bokeh.models import GlyphRenderer


@dataclass(frozen=True, kw_only=True)
class Wireframe:
    """
    Base class for all wireframes. Wireframe is a set of arguments
    that are passed to build a plot. Then it has the logic of
    transforming these arguments into the plot itself, including
    building the underlying dataset.
    """

    numpoints: int
    x_axis_name: str | None = None
    y_axis_name: str | None = None

    @property
    def data_key(self):
        """
        The underlying dataset is considered the same for the same data_key.
        For example it's good to have the same datakey for the same ERC20 token.
        """

    @property
    def x_axis(self):
        """
        The bokeh type for the x axis
        """
        raise RuntimeError("Not implemented")

    @property
    def y_axis(self):
        """
        The name of the axis. Data with the same axis name will
        be plotted on the same axis.
        """
        raise RuntimeError("Not implemented")

    def x(self, data: Any) -> List[Any]:
        """
        A list of x values for the plot.

        Arguments:
            data: Underlying dataset

        Returns:
            A list of x values
        """
        raise RuntimeError("Not implemented")

    def y(self, data: Any) -> List[Any]:
        """
        A list of y values for the plot.

        Arguments:
            data: Underlying dataset

        Returns:
            A list of y values
        """
        raise RuntimeError("Not implemented")

    def build_data(self, default_data: Any | None, **core_args) -> Any:
        """
        Build data from the wireframe arguments.

        Arguments:
            default_data: A data with the same data_key that was built
                          from other wireframes.

        Returns:
            An updated or created dataset.
        """
        raise RuntimeError("Not implemented")

    def plot(self, fig: Figure, x: Any, y: Any, **kwargs) -> GlyphRenderer:
        """
        Make a plot on figure.

        Arguments:
            fig: :class:`bokeh.plotting.Figure` for the plot
            x: A list of x values
            y: A list of y values

        Returns:
            A plot
        """
        raise RuntimeError("Not implemented")


@dataclass(frozen=True)
class TimeseriesWireframe(Wireframe):
    """
    An abstract class for any timeseries wireframe
    """

    #: Start of the timeseries
    start: int | datetime
    #: End of the timeseries
    end: int | datetime

    @property
    def x_axis(self) -> str:
        return "datetime"

    def x(self, data: Any):
        start = self._resolve_datetime(self.start)
        end = self._resolve_datetime(self.end)
        step = (end - start) // (self.numpoints - 1)
        timestamps = [*range(start, end, step)]
        if timestamps[-1] != end:
            timestamps.append(end)
        dates = [datetime.fromtimestamp(t) for t in timestamps]
        return dates

    def _resolve_datetime(self, tim: int | datetime):
        if isinstance(tim, datetime):
            return int(time.mktime(tim.timetuple()))
        return tim



from __future__ import annotations
from typing import Any, Dict, Tuple
import json


class Event:
    """
    Event represents an event log on Ethereum blockchain.
    """

    #: Ethereum chain_id
    chain_id: int
    #: The block this event appeared in
    block_number: int
    #: The hash of the transaction this event appeared in
    transaction_hash: str
    #: The log number for this event inside the transaction
    log_index: int
    #: The contract address this event appeared in (lowercase)
    address: str
    #: Event name
    event: str
    #: Event arguments (hexes are lowercase)
    args: Dict[str, Any]

    def __init__(
        self,
        chain_id: int,
        block_number: int,
        transaction_hash: str,
        log_index: int,
        address: str,
        event: str,
        args: Dict[str, Any],
    ):
        self.chain_id = chain_id
        self.block_number = block_number
        self.transaction_hash = transaction_hash
        self.log_index = log_index
        self.address = address.lower()
        self.event = event
        self.args = self._lower(args)

    @staticmethod
    def from_row(row: Tuple[int, int, str, int, str, str, str]) -> Event:
        """
        Deserialize from web3cat.database row

        Args:
            row: database row
        """
        event = Event(*row)
        event.args = json.loads(event.args)
        return event

    def to_row(self) -> Tuple[int, int, str, int, str, str, str]:
        """
        Serialize to database row

        Returns:
            database row
        """
        return (
            self.chain_id,
            self.block_number,
            self.transaction_hash,
            self.log_index,
            self.address,
            self.event,
            json.dumps(self.args),
        )

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert :class:`Event` to dict
        """
        return {
            "chainId": self.chain_id,
            "blockNumber": self.block_number,
            "transactionHash": self.transaction_hash,
            "logIndex": self.log_index,
            "address": self.address,
            "event": self.event,
            "args": self.args,
        }

    @staticmethod
    def from_dict(dct: Dict[str, Any]):
        """
        Create :class:`Event` from dict
        """
        return Event(
            chain_id=dct["chainId"],
            block_number=dct["blockNumber"],
            transaction_hash=dct["transactionHash"],
            log_index=dct["logIndex"],
            address=dct["address"].lower(),
            event=dct["event"],
            args=Event._lower(dct["args"]),
        )

    def matches_filter(self, event_filter: Dict[str, Any] | None) -> bool:
        """
        Checks if :class:`Event` matches given event event_filter.

        Args:
            event_filter: Event event_filter

        Returns:
            ``True`` if matches, ``False`` otherwise
        """
        if event_filter is None or event_filter == {}:
            return True
        if self.args is None:
            return False
        for k in event_filter.keys():
            if not k in self.args:
                return False
            if not self._value_match_filter(self.args[k], event_filter[k]):
                return False
        return True

    def _value_match_filter(self, value, filter_value):
        # the most basic case: 2 plain values
        if not isinstance(filter_value, list) and not isinstance(value, list):
            return value == filter_value
        # filter_value is a list of possible values (OR filter) and value is list
        if isinstance(filter_value, list) and not isinstance(value, list):
            for ifv in filter_value:
                if ifv == value:
                    return True
        # filter value is plain value but value is list
        if not isinstance(filter_value, list) and isinstance(value, list):
            return False
        # Now we have both values as lists
        # Case 1: filter_value is []. It is a plain list comparison then.
        # Doesn't make sense to supply [] as an empty list of ORs
        if len(filter_value) == 0:
            return value == filter_value

        # Case 2: filter_value is a list of lists. Then it's OR on lists
        if isinstance(filter_value[0], list):
            for fv in filter_value:
                if fv == value:
                    return True
        # Case 3: filter_value is a list and value is a list
        return value == filter_value

    @classmethod
    def _lower(cls, val: Any) -> Dict[str, Any]:
        if val is None:
            return None
        if isinstance(val, list):
            for i, v in enumerate(val):
                val[i] = cls._lower(v)
        if isinstance(val, dict):
            for k in val.keys():
                val[k] = cls._lower(val[k])
        if isinstance(val, str) and val.startswith("0x"):
            val = val.lower()
        return val

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False

    def __repr__(self):
        return f"Event({json.dumps(self.to_dict())})"



from dataclasses import dataclass
from typing import List
from functools import cached_property
from datetime import datetime
from bokeh.plotting import Figure
from bokeh.models import GlyphRenderer
import numpy as np


from web3cat.fetcher.utils import short_address
from web3cat.fetcher.erc20_metas import ERC20Meta
from web3cat.data.erc20s import ERC20Data
from web3cat.view.wireframes.abstract import TimeseriesWireframe


@dataclass(frozen=True)
class TotalSupplyWireframe(TimeseriesWireframe):
    """
    ERC20 total supply wireframe
    """

    token: ERC20Meta

    @cached_property
    def data_key(self) -> str:
        return f"erc20_{self.token.symbol.upper()}"

    @property
    def y_axis(self) -> str:
        return f"Total Supply ({self.token.symbol.upper()})"

    def build_data(self, default_data: ERC20Data | None, **core_args) -> ERC20Data:
        data = ERC20Data(
            token=self.token.address,
            address_filter=[],
            start=self.start,
            end=self.end,
            **core_args,
        )
        if default_data is None:
            return data
        return ERC20Data(
            token=self.token.address,
            address_filter=default_data.address_filter,
            start=min(data.from_block_number, default_data.from_block_number),
            end=max(data.to_block_number, default_data.to_block_number),
        )

    def y(self, data: ERC20Data) -> List[np.float64]:
        return data.total_supply(self.x(data))["total_supply"].to_list()

    def plot(
        self, fig: Figure, x: List[datetime], y: List[np.float64], **kwargs
    ) -> GlyphRenderer:
        return fig.line(
            x,
            y,
            line_width=2,
            legend_label=self.y_axis,
            **kwargs,
        )


@dataclass(frozen=True)
class BalanceWireframe(TimeseriesWireframe):
    """
    ERC20 balances wireframe
    """

    token: ERC20Meta
    address: str

    @cached_property
    def data_key(self) -> str:
        return f"erc20_{self.token.symbol.upper()}"

    @property
    def y_axis(self) -> str:
        return f"Balance ({self.token.symbol.upper()})"

    def build_data(self, default_data: ERC20Data | None, **core_args) -> ERC20Data:
        data = ERC20Data(
            token=self.token.address,
            address_filter=[self.address],
            start=self.start,
            end=self.end,
            **core_args,
        )
        if default_data is None:
            return data
        return ERC20Data(
            token=self.token.address,
            address_filter=default_data.address_filter + [self.address],
            start=min(data.from_block_number, default_data.from_block_number),
            end=max(data.to_block_number, default_data.to_block_number),
        )

    def y(self, data: ERC20Data) -> List[np.float64]:
        return data.balances([self.address], self.x(data))["balance"].to_list()

    def plot(
        self, fig: Figure, x: List[datetime], y: List[np.float64], **kwargs
    ) -> GlyphRenderer:
        return fig.line(
            x,
            y,
            line_width=2,
            legend_label=f"{short_address(self.address.lower())} balance"
            f" ({self.token.symbol.upper()})",
            **kwargs,
        )



from __future__ import annotations
import json
from typing import Any, Dict, Tuple
from web3.contract import Contract


class ERC20Meta:
    """
    ERC20 token metadata (name, symbol, decimals).

    Note:
        The convention is to use ``address``, ``name`` and ``symbol`` in lowercase format.
        This is not in line with EIP55 but makes things more uniform and
        simpler.
    """

    #: Ethereum chain_id
    chain_id: int
    #: Token address (lowercase)
    address: str
    #: Token name (lowercase)
    name: str
    #: Token symbol (lowercase)
    symbol: str
    #: Token decimals
    decimals: int
    #: :class:`web3.Contract` for the token
    contract: Contract

    def __init__(
        self,
        chain_id: int,
        address: str,
        name: str,
        symbol: str,
        decimals: int,
        contract: Contract,
    ):
        self.chain_id = chain_id
        self.address = address.lower()
        self.name = name.lower()
        self.symbol = symbol.lower()
        self.decimals = decimals
        self.contract = contract

    @staticmethod
    def from_row(row: Tuple[int, str, str, str, int]) -> ERC20Meta:
        """
        Deserialize from web3cat.database row

        Args:
            row: database row
        """
        return ERC20Meta(contract=None, *row)

    def to_row(self) -> Tuple[int, str, str, str, int]:
        """
        Serialize to database row

        Returns:
            database row
        """

        return (self.chain_id, self.address, self.name, self.symbol, self.decimals)

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert :class:`ERC20Meta` to dict
        """
        return {
            "chainId": self.chain_id,
            "address": self.address,
            "name": self.name,
            "symbol": self.symbol,
            "decimals": self.decimals,
        }

    @staticmethod
    def from_dict(dct: Dict[str, Any]):
        """
        Create :class:`Call` from dict
        """

        return ERC20Meta(
            chain_id=dct["chainId"],
            address=dct["address"],
            name=dct["name"],
            symbol=dct["symbol"],
            decimals=dct["decimals"],
            contract=None,
        )

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False

    def __repr__(self):
        return f"ERC20Meta({json.dumps(self.to_dict())})"



from dataclasses import dataclass
from typing import List
from functools import cached_property
from datetime import datetime
from bokeh.plotting import Figure
from bokeh.models import GlyphRenderer
import numpy as np


from web3cat.fetcher.utils import short_address
from web3cat.data.ethers import EtherData
from web3cat.view.wireframes.abstract import TimeseriesWireframe


@dataclass(frozen=True)
class EthBalanceWireframe(TimeseriesWireframe):
    """
    Ether balances wireframe
    """

    address: str

    @cached_property
    def data_key(self) -> str:
        return "eth_balance"

    @property
    def y_axis(self) -> str:
        return "Balance (ETH)"

    def build_data(self, default_data: EtherData | None, **core_args) -> EtherData:
        if not default_data is None:
            return default_data
        return EtherData(
            start=self.start,
            end=self.end,
            **core_args,
        )

    def y(self, data: EtherData) -> List[np.float64]:
        return data.balances([self.address], self.x(data))["balance"].to_list()

    def plot(
        self, fig: Figure, x: List[datetime], y: List[np.float64], **kwargs
    ) -> GlyphRenderer:
        return fig.line(
            x,
            y,
            line_width=2,
            legend_label=f"{short_address(self.address.lower())} balance (ETH)",
            **kwargs,
        )



from __future__ import annotations

import json
from typing import Dict, List, Literal
from web3.exceptions import BlockNotFound

from web3cat.fetcher.blocks.repo import BlocksRepo
from web3cat.fetcher.blocks.block import Block
from web3cat.fetcher.core import Core
from web3cat.fetcher.utils import json_response, print_progress


class BlocksService(Core):
    """
    Service for fetching and caching Ethereum block data.

    **Request/Response flow**

    ::

                +---------------+            +-------+ +-------------+
                | BlocksService |            | Web3  | | BlocksRepo  |
                +---------------+            +-------+ +-------------+
        -----------------  |                        |            |
        | Request blocks |-|                        |            |
        |----------------| |                        |            |
                           |                        |            |
                           | Get blocks             |            |
                           |------------------------------------>|
                           |                        |            |
                           | Get missing blocks     |            |
                           |----------------------->|            |
                           |                        |            |
                           | Save missing blocks    |            |
                           |------------------------------------>|
              -----------  |                        |            |
              | Response |-|                        |            |
              |----------| |                        |            |
                           |                        |            |

    Args:
        blocks_repo: An instance of :class:`fetcher.blocks.BlocksRepo`
        kwargs: Args for the :class:`fetcher.core.Core`

    See Also:
        :class:`fetcher.core.Core` for defining the block grid and
        how timestamps are approximated.

    """

    _blocks_repo: BlocksRepo
    _latest_block: Block | None
    _block_cache: Dict[int, Block]
    _block_grid_cache: Dict[int, Block]

    def __init__(self, blocks_repo: BlocksRepo, **kwargs):
        super().__init__(**kwargs)
        self._blocks_repo = blocks_repo
        self._latest_block = None
        self._block_cache = {}
        self._block_grid_cache = {}

    @staticmethod
    def create(**kwargs) -> BlocksService:
        """
        Create an instance of :class:`BlocksService`

        Args:
            kwargs: Args for the :class:`fetcher.core.Core`

        Returns:
            An instance of :class:`BlocksService`
        """
        blocks_repo = BlocksRepo(**kwargs)
        service = BlocksService(blocks_repo)
        return service

    @property
    def latest_block(self) -> Block:
        """
        Latest block from Ethereum (this value is cached on the first call)
        """
        if self._latest_block is None:
            self._latest_block = self._fetch_block_from_rpc()
        return self._latest_block

    def refresh_latest_block(self):
        """
        Refresh :attr:`latest_block` (the value is cached by default)
        """
        self._latest_block = self._fetch_block_from_rpc()

    def get_latest_block_at_timestamp(self, timestamp: int) -> Block | None:
        """
        Get the first block after a timestamp.

        Args:
            timestamp: UNIX timestamp

        Returns:
            First block after timestamp, ``None`` if the block doesn't exist
        """
        left_block = self._blocks_repo.get_block_before_timestamp(timestamp)
        if left_block is None:
            left_block = self._get_grid_block(1)

        right_block = self._blocks_repo.get_block_after_timestamp(timestamp)
        if right_block is None:
            right_block = self.latest_block

        if timestamp >= right_block.timestamp:
            return self.latest_block
        if timestamp < left_block.timestamp:
            return None

        # invariant: left_block.timestamp <= timestamp < right_block.timestamp
        while right_block.number - left_block.number > self.block_grid_step:
            w = (timestamp - left_block.timestamp) / (
                right_block.timestamp - left_block.timestamp
            )
            num = self._snap_to_grid(
                int((1 - w) * left_block.number + w * right_block.number),
                direction="left",
            )
            # out block somewhere close to left_block
            if num == left_block.number:
                num = self._next_grid_block(num)

            block = self._get_grid_block(num)
            if block.timestamp > timestamp:
                right_block = block
            else:
                left_block = block

        return self._synthesize_block_from_timestamp(left_block, right_block, timestamp)

    def get_latest_blocks_by_timestamps(
        self, block_timestamps: int | List[int]
    ) -> List[Block]:
        """
        Get a list of latest blocks as of timestamp.

        Args:
            block_timestamps: Unix timestamps

        Returns:
            A list of blocks in the same order
        """
        if not isinstance(block_timestamps, list):
            block_timestamps = [block_timestamps]

        if len(block_timestamps) == 0:
            return []
        block_idx = {}
        timestamps_to_fetch = []
        for timestamp in block_timestamps:
            block = self._resolve_from_cache(timestamp)
            if block is None:
                timestamps_to_fetch.append(timestamp)
            else:
                block_idx[timestamp] = block
        for i, timestamp in enumerate(timestamps_to_fetch):
            print_progress(
                i,
                len(timestamps_to_fetch),
                f"Resolving {len(timestamps_to_fetch)} block numbers",
            )
            block_idx[timestamp] = self.get_latest_block_at_timestamp(timestamp)

        if len(timestamps_to_fetch) > 0:
            print_progress(
                len(timestamps_to_fetch),
                len(timestamps_to_fetch),
                f"Resolving {len(timestamps_to_fetch)} block numbers",
            )

        return [block_idx[timestamp] for timestamp in block_timestamps]

    def get_blocks(self, numbers: int | List[int]) -> List[Block]:
        """
        Get blocks by numbers.

        Args:
            number: block numbers

        Returns:
            Blocks with these numbers. ``None`` if the block doesn't exist.
        """

        if not isinstance(numbers, list):
            numbers = [numbers]

        if len(numbers) == 0:
            return []

        out = []

        block_numbers = list(set(numbers))
        grid_block_numbers = []

        for b in block_numbers:
            snapped = self._snap_to_grid(b, direction="left")
            grid_block_numbers.append(snapped)
            if snapped != b:
                right = self._next_grid_block(snapped)
                if not right is None:
                    grid_block_numbers.append(right)

        grid_block_numbers = list(set(grid_block_numbers))
        cached_grid_blocks = self._blocks_repo.find(grid_block_numbers)
        grid_block_idx = {b.number: b for b in cached_grid_blocks}
        grid_block_numbers_to_fecth = []
        for num in grid_block_numbers:
            if not num in grid_block_idx:
                grid_block_numbers_to_fecth.append(num)

        for i, num in enumerate(grid_block_numbers_to_fecth):
            print_progress(
                i,
                len(grid_block_numbers_to_fecth),
                f"Resolving {len(grid_block_numbers_to_fecth)} block numbers",
            )
            block = self._fetch_block_from_rpc_and_save(num)
            grid_block_idx[block.number] = block

        if len(grid_block_numbers_to_fecth) > 0:
            print_progress(
                len(grid_block_numbers_to_fecth),
                len(grid_block_numbers_to_fecth),
                f"Resolving {len(grid_block_numbers_to_fecth)} block numbers",
            )

        out = []
        for num in numbers:
            left_num = self._snap_to_grid(num, direction="left")
            if left_num == num:
                out.append(grid_block_idx[num])
                continue

            right_num = self._snap_to_grid(num, direction="right")
            left = grid_block_idx[left_num]
            right = grid_block_idx[right_num]
            block = self._synthesize_block(left, right, num)
            out.append(block)

        return out

    def clear_cache(self):
        """
        Delete all cached entries
        """
        self._block_cache = {}
        self._block_grid_cache = {}
        self._blocks_repo.purge()
        self._blocks_repo.conn.commit()

    def _resolve_from_cache(self, timestamp: int) -> Block | None:
        if timestamp in self._block_cache:
            return self._block_cache[timestamp]

        left = self._blocks_repo.get_block_before_timestamp(timestamp)
        right = self._blocks_repo.get_block_after_timestamp(timestamp)
        if left is None or right is None:
            return None
        if right.number - left.number == self.block_grid_step:
            return self._synthesize_block_from_timestamp(left, right, timestamp)
        return None

    def _synthesize_block(
        self,
        left_block: Block,
        right_block: Block,
        number: int,
    ) -> Block:
        if left_block == right_block:
            return Block(self.chain_id, left_block.number, left_block.timestamp)
        if number == left_block.number:
            return left_block
        if number == right_block.number:
            return right_block
        w = (number - left_block.number) / (right_block.number - left_block.number)
        timestamp = int((1 - w) * left_block.timestamp + w * right_block.timestamp)
        block = Block(self.chain_id, number, timestamp)
        self._block_cache[timestamp] = block
        return block

    def _synthesize_block_from_timestamp(
        self,
        left_block: Block,
        right_block: Block,
        timestamp: int,
    ) -> Block:
        if left_block == right_block:
            return Block(self.chain_id, left_block.number, left_block.timestamp)
        w = (timestamp - left_block.timestamp) / (
            right_block.timestamp - left_block.timestamp
        )
        number = int((1 - w) * left_block.number + w * right_block.number)
        prev = self._synthesize_block(left_block, right_block, number)
        nxt = self._synthesize_block(left_block, right_block, number + 1)
        if nxt.timestamp > timestamp:
            return prev
        return nxt

    def _snap_to_grid(
        self, block_number: int, direction=Literal["left"] | Literal["right"]
    ) -> int:
        if block_number >= self.latest_block.number:
            return block_number
        if block_number <= 1:
            # this doesn't fit in the logic below because of the small grid_step
            if block_number == 1 and self.block_grid_step == 1 and direction == "right":
                return 2
            return 1
        mod = block_number % self.block_grid_step
        snapped = block_number
        if mod != 0:
            if direction == "left":
                snapped = block_number - mod
            else:
                snapped = block_number - mod + self.block_grid_step

        # the lower bound for the grid is block 1
        if snapped < 1:
            return 1
        # the upper bound for the grid is the latest block
        if snapped > self.latest_block.number:
            return self.latest_block.number
        return snapped

    def _next_grid_block(self, num: int) -> int | None:
        if num == self.latest_block.number:
            return None
        if self.block_grid_step == 1:
            return num + 1

        return self._snap_to_grid(num + 1, direction="right")

    def _get_grid_block(self, number: int | None) -> Block:
        if number is None:
            return self.latest_block
        snapped_num = self._snap_to_grid(number, direction="left")
        if number != snapped_num:
            raise ValueError("API call for blocks out of grid are prohibited")

        if number in self._block_grid_cache:
            return self._block_grid_cache[number]

        block = next(iter(self._blocks_repo.find(number)), None)
        if not block is None:
            self._block_grid_cache[number] = block
            return block

        return self._fetch_block_from_rpc_and_save(number)

    def _fetch_block_from_rpc_and_save(self, number: int) -> Block | None:
        block = self._fetch_block_from_rpc(number)
        if block is None:
            if number < 1:
                block = self._fetch_block_from_rpc(1)
            else:
                block = self.latest_block
        self._blocks_repo.save([block])
        self._blocks_repo.conn.commit()
        self._block_grid_cache[block.number] = block

        return block

    def _fetch_block_from_rpc(self, number: int | None = None) -> Block | None:
        block_id = "latest" if number is None else number
        try:
            raw_block = self.w3.eth.get_block(block_id)
            block = json.loads(json_response(raw_block))
            return Block.from_dict({"chainId": self.chain_id, **block})

        except BlockNotFound:
            return None



"""
Module for fetching and caching events from web3.

The main class of this module is :class:`EventsService`.
It is used for retrieving events from web3 and storing them in
cache so that subsequent requests are returned from cache.

It also supports quick incremental fetches. For example,
if you fetched ERC20 Transfer events from block 10_000 to
block 20_000, then a subsequent request from block 15_000 to
block 21_000 will return 15_000 - 20_000 from cache and fetch
only 20_000 - 21_000 from web3 and save to cache.
A subsequent request for blocks 15_000 - 21_000 will read
everything from cache.

Additionally, it supports nesting of the argument filters.
Imagine you

1. Fetched all Transfer events from blocks 2000 - 4000
2. Fetched Transfer events with filter :code:`{"from": "0x1234..."}` for blocks 4000 - 6000
3. Fetched Transfer events with filter :code:`{"to": "0x5678..."}` for blocks 6000 - 8000

Now when you query events with filter :code:`{"from": "0x1234...", "to": "0x5678..."}`
for blocks 2000 - 8000 :class:`EventsService` is smart enough
to figure out that all events are already in cache and
serve the cached result.

Example:
    ::

        from web3cat.fetcher.events import EventsService
        from web3cat.fetcher.erc20_metas import ERC20MetasService

        erc20 = ERC20MetasService.create()
        dai = erc20.get("Dai")

        service = EventsService.create()
        events = service.get_events(
            dai.contract.events.Transfer, from_block=15830000, to_block=15840100
        )
"""

from web3cat.fetcher.events.event import Event
from web3cat.fetcher.events.repo import EventsRepo
from web3cat.fetcher.events.service import EventsService



from web3cat.view.wireframes.abstract import *
from web3cat.view.wireframes.erc20 import *
from web3cat.view.wireframes.chainlink import *
from web3cat.view.wireframes.eth import *
from web3cat.view.wireframes.portfolio import *



from __future__ import annotations
import json
import os
from typing import Any, Dict
from web3.contract import Contract
from web3cat.fetcher.core import Core

from web3cat.fetcher.erc20_metas.erc20_meta import ERC20Meta
from web3cat.fetcher.erc20_metas.repo import ERC20MetasRepo


class ERC20MetasService(Core):
    """
    Service for fetching ERC20 tokens metadata (name, symbol, decimals).

    The sole purpose of this service is to fetch ERC20 tokens metadata from web3,
    cache it, and read from the cache on subsequent calls.

    The exact flow goes like this

    **Request/Response flow**

    ::

                    +-------------------+                     +-------+ +---------------+ +-----------------+
                    | ERC20MetasService |                     | Web3  | | PreloadedData | | ERC20MetasRepo  |
                    +-------------------+                     +-------+ +---------------+ +-----------------+
         -------------------  |                                   |             |                  |
         | Metadata request |-|                                   |             |                  |
         |------------------| |                                   |             |                  |
                              |                                   |             |                  |
                              | Find metadata                     |             |                  |
                              |------------------------------------------------>|                  |
                              |                                   |             |                  |
                              | If cache miss: Find metadata      |             |                  |
                              |------------------------------------------------------------------->|
                              |                                   |             |                  |
                              | If cache miss: Fetch metadata     |             |                  |
                              |---------------------------------->|             |                  |
                              |                                   |             |                  |
                              | Save metadata                     |             |                  |
                              |------------------------------------------------------------------->|
        --------------------  |                                   |             |                  |
        | Metadata response |-|                                   |             |                  |
        |-------------------| |                                   |             |                  |
                              |                                   |             |                  |

    Args:
        erc20_metas_repo: :class:`ERC20MetasRepo` instance
        kwargs: Args for the :class:`fetcher.core.Core`
    """

    _cache: Dict[str, Any]
    _erc20_abi: Dict[str, Any]
    _erc20_metas_repo: ERC20MetasRepo

    def __init__(self, erc20_metas_repo: ERC20MetasRepo, **kwargs):
        super().__init__(**kwargs)
        current_folder = os.path.realpath(os.path.dirname(__file__))
        with open(f"{current_folder}/tokens.json", "r", encoding="utf-8") as f:
            self._cache = json.load(f)
        with open(f"{current_folder}/erc20_abi.json", "r", encoding="utf-8") as f:
            self._erc20_abi = json.load(f)
        self._erc20_metas_repo = erc20_metas_repo

    @staticmethod
    def create(**kwargs) -> ERC20MetasService:
        """
        Create an instance of :class:`ERC20MetasService`

        Args:
            kwargs: Args for the :class:`fetcher.core.Core`

        Returns:
            An instance of :class:`ERC20MetasService`
        """

        erc20_metas_repo = ERC20MetasRepo(**kwargs)
        return ERC20MetasService(erc20_metas_repo, **kwargs)

    def get(self, token: str) -> ERC20Meta | None:
        """
        Get metadata by token symbol or token address.

        Getting token metadata by symbol only works for cached metadata.
        The preloaded cache is large and contains major
        tokens for all evm chains.

        Args:
            token: token symbol or token address

        Returns:
            An instance of :class:`ERC20Meta` or ``None`` if not found.
        """
        token = token.lower()
        cached_token = self._get_from_cache(token)
        if not cached_token:
            cached_token = self._erc20_metas_repo.find(token)
        if cached_token:
            contract: Contract = self.w3.eth.contract(
                address=self.w3.to_checksum_address(cached_token.address),
                abi=self._erc20_abi,
            )
            cached_token.contract = contract
            return cached_token
        if not token.startswith("0x"):
            raise ValueError(f"Could not find token `{token}`")
        contract: Contract = self.w3.eth.contract(
            address=self.w3.to_checksum_address(token), abi=self._erc20_abi
        )
        decimals = contract.functions.decimals().call()
        name = contract.functions.name().call()
        symbol = contract.functions.symbol().call()
        meta = ERC20Meta(
            self.chain_id, token, name, symbol, decimals, contract=contract
        )
        self._erc20_metas_repo.save([meta])
        self._erc20_metas_repo.conn.commit()
        return meta

    def clear_cache(self):
        """
        Delete all cached entries
        """
        self._erc20_metas_repo.purge()
        self._erc20_metas_repo.conn.commit()

    def _get_from_cache(self, token: str) -> ERC20Meta | None:
        token = token.lower()
        chain_id = str(self.chain_id)
        if not chain_id in self._cache:
            return None
        if not token in self._cache[chain_id]:
            return None
        data = self._cache[chain_id][token]
        return ERC20Meta(
            self.chain_id,
            data["address"],
            data["name"],
            data["symbol"],
            data["decimals"],
            None,
        )
